GfxAddressCheck equ 0

section .text
;
;if first bit set then
;plotpixel:
;  if second bit set then
;    use second color
;  else
;    use first color
;else
;  end row if no more opaque pixels
;  loop until next opaque pixel
;  goto plotpixel
;next row
;
; (
;  esi=font character source
;  edi=pixel destination
;  al/ah=first/second color
;  ecx=height/width/preshift (none will be greater than 255)
;  edx=wrap (destination width)
; )
;
BlitTwoColorFont:
  %if GfxAddressCheck
    cmp edi,Screen.Size
    jb .Ok
    int3
  .Ok:
  %endif
    add edi,[Display.Ptr]
.AbsoluteAddress:
    push ebp                    ;assume width=8 (ch) and display=320 (edx)
    movzx ebp,ch                ;8
    sub ebp,edx                 ;8 - 320
    mov edx,ecx                 ;copy height and width
    shr ecx,16                  ;bring preshift down to cl
    neg ebp                     ;-312 -> 312
    shl cl,1                    ;preshift*2 since each pixel is two bits

    ;esi=font pixel source
    ;edi=pixel destination
    ;al =first color   ah=second color
    ;ebx=font pixel data
    ;cl =preshift      ch=column
    ;edx=counters
    ;ebp=wrap

.NextRow:
    mov ch,dh                   ;reset column counter
    mov ebx,[esi]
    shr ebx,cl                  ;preshift for clipping
.NextPixel:
    ;if first bit is clear then skip transparent pixel
    shr ebx,1
    jnc .ClearPixel
.CheckPixelColor:
    ;if second bit is clear then use first color, if set then use second
    shr ebx,1
    jc .SecondPixelColor
    mov [edi],al                ;write first color
    inc edi
    dec ch                      ;one less column
    jnz .NextPixel
    jmp short .EndRow
.SecondPixelColor:
    mov [edi],ah                ;write second color
    inc edi
    dec ch                      ;one less column
    jnz .NextPixel
    jmp short .EndRow
.ClearPixel:
    ;enter a tight loop if more opaque pixels remain from the above shift
    jnz .ClearPixelLoop
    ;end row if there are only more transparent pixels afterwards
    movzx ebx,ch
    add edi,ebx
    jmp .EndRow
    ;loop until opaque pixel found
.NextClearPixel:
    shr ebx,2
    jc .CheckPixelColor
.ClearPixelLoop:
    inc edi
    dec ch                      ;one less column
    jnz .NextClearPixel

.EndRow:
    add edi,ebp
    add esi,byte 4
    dec dl                      ;one less row
    jnz .NextRow

    pop ebp
    ret


;
; The code for clipping is contained in here rather than in the individual
; character blitter each time it is called.
;
; (words row, column, string ptr, dwords string length)
; (esi=character just after last one blit)
BlitString:
    call GetFontClips
    jc .FullClip
.AlreadyClipped:
    mov esi,[esp+8]             ;get string ptr
    mov edi,[ClippedGfx.Ptr]
    push dword [esp+12]         ;save text length counter
    jmp short .FirstChar        ;in case length is zero characters
.FullClip:
    ret

.NextChar:
    push edi
    push esi
    movzx eax,byte [esi]
    mov ebx,[Font.SizesPtr]
    mov esi,[ClippedGfx.CharsPtr];get graphics base of character set
    movzx ecx,byte [ebx+eax]    ;get width of character
    shl eax,5                   ;*32
    inc ecx                     ;move right by character size plus one column
    add esi,eax                 ;get ptr to character bitmap
    add [esp+4],ecx             ;add character width to destination

    ;if ptr greater than right clip
    ;  clip width
    ;  skip if width < 1
    ;endif
    ;if ptr less than left clip
    ;  clip width
    ;  skip if width < 1
    ;  get preshift
    ;  add preshift to destination
    ;endif

    cmp [ClippedGfx.RightPtr],edi;if current ptr > end right clip
    jge .RightOk
    mov ebx,[ClippedGfx.RightPtr]
    add ebx,[ClippedGfx.Width]
    sub ebx,edi                 ;maximum width -= (right ptr - current ptr + char width)
    jle .Skip
    cmp ebx,ecx                 ;leave width alone if it is already smaller
    jae .RightOk
    mov ecx,ebx                 ;otherwise clip character width
.RightOk:

    cmp [ClippedGfx.LeftPtr],edi;if current ptr < left clip
    jle .LeftOk
    mov ebx,[ClippedGfx.LeftPtr];get preshift
    sub ebx,edi                 ;preshift = left ptr - current ptr
    sub ecx,ebx                 ;width -= (left ptr - current ptr)
    jbe .Skip
    add edi,ebx                 ;destination += preshift
    mov ch,bl                   ;move preshift into height/width dword
.LeftOk:

    mov eax,[Font.Colors]
    shl ecx,8                   ;shift width and preshift
    mov edx,[Display.Width]
    mov cl,[ClippedGfx.Height]  ;get height of character
    call BlitTwoColorFont
.Skip:
    pop esi
    pop edi
    inc esi                     ;next character
.FirstChar:
    dec dword [esp]             ;decrement string length counter
    jns .NextChar
    add esp,byte 4              ;release counter

    mov [ClippedGfx.Ptr],edi
    mov [esp+8],esi
.End:
    ret


;
; Draws a singled colored box.
;
; (TopRow, LeftCol, Height, Width, Color)
DrawRect:
%if 0
    call GetRectClips
    jc .FullClip
    ;mov ebx,[ClippedGfx.Width-2]
    shl ebx,16
    mov bx,[ClippedGfx.Height]
    ;mov edi,[ClippedGfx.Ptr]
    ;mov edx,[ClippedGfx.Wrap]
%endif

.NoClip:
    movzx ecx,word [esp+4]      ;get toprow
    movzx eax,word [esp+6]      ;get leftcol
    movzx edx,word [esp+10]     ;get width
    mov edi,[Display.Width]     ;get screen width
    neg edx
    add edx,edi                 ;get wrap width (screen width - box width)
    imul edi,ecx                ;row * screen width
    mov ebx,[esp+8]             ;get height and width (hl=width:height)
    add edi,eax                 ;(row * screen width) + left column
    ;jmp short .Start
.Start:
  %if GfxAddressCheck
    cmp edi,Screen.Size
    jb .Ok
    int3
  .Ok:
  %endif
    add edi,[Display.Ptr]
    mov al,[esp+12]             ;get color

;(al=color, ebx=height/width, edi=destination, edx=wrap/row sep)
.ByReg:                         ;calling by register can be a bit faster
    cld
    mov ah,al                   ;copy color to second byte
    mov esi,edi                 ;copy destination
    shrd ecx,eax,16
    neg esi
    shld eax,ecx,16
	;eax now consists of four pixels all the same color
    xor ecx,ecx                 ;zero top bits for loop
    and esi,3                   ;get bytes to next higher dword
    shl esi,16                  ;move to upper word
    sub ebx,esi                 ;remaining width = width - first strip length
    jae .WidthPositive
    add ebx,esi                 ;reverse subtraction above
    xor esi,esi
.WidthPositive:
    shr esi,16                  ;return to lower word

.NextLine:
    mov ecx,esi                 ;(-destination) % 4
    rep stosb
    shld ecx,ebx,14             ;(remaining width) \ 4
    rep stosd
    shld ecx,ebx,16
    and ecx,3                   ;(remaining width) % 4
    rep stosb
	add edi,edx
	dec bx
	jnz .NextLine
.End:
	ret

%if 0
.FullClip:
    stc
    ret
%endif


;
;(esi=palette)
SetPalette:
    mov ecx,256
;(esi=palette, ecx=count)
.GivenCount:
    xor eax,eax
;(esi=palette, eax=starting reg, ecx=count)
.GivenIndexes:
    mov edx,3C8h
    out dx,al

    inc edx
    and ecx,511
    lea ecx,[ecx*2+ecx]
    cld
    rep outsb
    ret


;
; (words row, column)
; (cf=full clip, edi=destination)
GetFontClips:
    movsx eax,word [esp+8]      ; get top row
    movzx ecx,byte [Font.Height]; get character height
    mov esi,[Font.CharsPtr]
    mov edi,[Display.Width]
    mov ebx,eax                 ; copy top row
.CheckTop:
    cmp [Display.TopClip],eax   ; if top row < top clip
    jle .TopOk                  ;   difference = bottom clip - bottom column
    mov edx,[Display.TopClip]   ;   set top row to top clip
    mov eax,edx                 ;   *4 bytes per character row
    sub edx,ebx                 ; endif
    lea esi,[esi+edx*4]
.TopOk:
.CheckBtm:
    add ebx,ecx                 ; bottom row = top row + height
    cmp [Display.BtmClip],ebx   ; if bottom row > bottom clip
    jge .BtmOk                  ;   set bottom row to bottom clip
    mov ebx,[Display.BtmClip]   ; endif
.BtmOk:
    sub ebx,eax                 ; clipped height = bottom side - top side
    jle .FullClip               ; exit if clipped height <= 0

    imul edi,eax                ; destination = top row * screen width
    movzx ecx,byte [Font.Width] ; get character width
    mov [ClippedGfx.CharsPtr],esi
    mov [ClippedGfx.Height],ebx ; set character height
    mov [ClippedGfx.Width],ecx  ; set character width

    movsx eax,word [esp+10]     ; get left column
    mov ebx,[Display.LeftClip]
    mov edx,[Display.RightClip]
    sub edx,ecx                 ; right clip - char width
    add ebx,edi                 ; left clip + row ptr
    add edx,edi                 ; (right clip - char width) + row ptr
    add edi,eax                 ; destination = row ptr + left column
    mov [ClippedGfx.LeftPtr],ebx
    mov [ClippedGfx.RightPtr],edx
    mov [ClippedGfx.Ptr],edi

    clc
    ret

.FullClip:
    stc
    ret


;
section .data

;Screen:
;.Height     equ 200
;.Width      equ 320
;.Ptr        equ 0A0000h
;.Size       equ .Height * .Width

align 4
Display:
.Height:    dd Screen.Height
.Width:     dd Screen.Width
.Wrap:      dd Screen.Width     ;bytes per row
.BasePtr:   dd Screen.Adr       ;actual base of display
.Ptr:       dd Screen.Adr       ;current offset relative to top/left
.Top:       dd 0                ;current window row offset
.Left:      dd 0                ;current window column offset
.TopClip:   dd 0                ;clip relative to Top
.BtmClip:   dd Screen.Height
.LeftClip:  dd 0                ;clip relative to Left
.RightClip: dd Screen.Width

ClippedGfx:
.Height:    dd 0
.Width:     dd 0
.TopAdjust: dd 0                ;offset from top after clip
.LeftAdjust:dd 0                ;offset from left after clip, example: a bitmap to be shown at column 50 is left clipped at 65, so the left offset is 15
.Ptr:       dd 0
.Wrap:      dd 0
.LeftPtr:   dd 0
.RightPtr:  dd 0
.CharsPtr:  dd 0

Font:
.Size:
.Height:    db 8                ;total height of character from top to bottom extension
.Width:     db 7
.BodyHeight:db 6                ;height of character's main body, not including the bottom extension found in "y","g","p","q"...
            db 0
.Colors:    db 4,2,0,0          ;fore color/border color/shadow color/background color
.CharsPtr:  dd GuiFontLetters
.SizesPtr:  dd GuiFontSizes
.CharacterSpacing   equ 1

alignb 4
GuiFontLetters equ $-(16*32)
    incbin "fnt_dflt.lbm"

GuiFontSizes   equ $-16
    db 6,5,5,5,5,6,5,5,5,4,5,5,5,5,5,5  ;special characters
    db 2,1,3,5,3,4,4,2,2,2,3,3,2,3,1,3  ;punctuation
    db 3,3,3,3,3,3,3,3,3,3,1,2,3,3,3,3  ;0-9
    db 4,3,3,3,3,3,3,3,3,1,3,3,3,5,3,3  ;A-O
    db 3,3,3,3,3,3,3,5,3,3,3,2,3,2,3,3  ;P-Z
    db 2,3,3,3,3,3,3,3,3,1,2,3,1,5,3,3  ;a-o
    db 3,3,3,2,3,3,3,5,3,3,3,3,1,3,3,3  ;p-z
