<SPC2MIDI.ASM>
;
SeekNoteNext:
    mov esi,[NoteBufferPos]
    mov eax,[TickTime]
    mov edx,[NoteBufferEnd]
    mov [NoteBufferPrevPos],esi
    jmp short .First
.Next:
    cmp [NoteBuffer+esi],eax    ;is time when note was recorded greater than current time
    ja .End
    add esi,byte 8
.First:
    cmp esi,edx                 ;check current position with end
    jb .Next
.End:
    mov [StSongPos.Time],eax
    mov [NoteBufferPos],esi
    ret


;
; Seeks a specified position in the Note Buffer, appropriately setting which
; notes are on/off, which channels have changed, and what each pitch is.
; This is mainly used for the note display modes, to simplify their code so
; they do not need to read the note buffer.
;
; (edi=seek structure, edx=new time to seek)
; (esi=new position)
;
SeekNote:
    ;start from given position
    ;loop until buffer pos >= buffer end or note time >= specified time
    ;  if note on
    ;    set note is on, flag note was turned on
    ;    set channel's instrument and pitch
    ;  elif note off
    ;    set note is off, flag note was turned off
    ;  endif
    ;endloop

    xor ecx,ecx  ;also needed for loop below
    mov esi,[edi+SeekNoteStruct.Pos]
    push dword ecx  ;keys turned off
    mov dword [edi+SeekNoteStruct.Change],ecx  ;clear any key changes
    jmp short .Start

    ; start loop
.Next:
    mov eax,[NoteBuffer+esi+4]

    ;   if note on
    ;     get voice/channel
    ;     turn on channel
    ;     copy attributes from buffer to current channel attributes
    ;     flag change in channel
    ;   else...
    test al,15
    mov cl,al
    jnz .OnEnd
    shr ecx,4
    bts [edi+SeekNoteStruct.Used],ecx
    mov [edi+SeekNoteStruct.Atrs+ecx*4],eax
    bts [edi+SeekNoteStruct.Change],ecx
    jmp short .ToNext
.OnEnd:

    ;   elif note off
    ;     get voice/channel
    ;     turn off channel
    ;     flag change in channel
    ;     if not already turned off once before
    ;       copy current channel attributes to previous attributes
    ;       copy current channel pitch to previous pitch
    ;     endif
    ;   endif
    test al,14
    jnz .OffEnd
    shr ecx,4  ;get voice/channel
    btr [edi+SeekNoteStruct.Used],ecx
    bts [esp],ecx  ;if channel wasn't already turned off once already
    jc .OffEnd
    bts [edi+SeekNoteStruct.Change],ecx
    mov eax,[edi+SeekNoteStruct.Atrs+ecx*4]
    mov [edi+SeekNoteStruct.PrevAtrs+ecx*4],eax
    mov bl,[edi+SeekNoteStruct.Pitches+ecx]
    mov [edi+SeekNoteStruct.PrevPitches+ecx],bl
    ;jmp short .ToNext
.OffEnd:

    ;   next note event in buffer
    ; endloop while not beyond buffer end and note time < desired time
.ToNext:
    add esi,byte NoteBuffer.eSize       ;next note in buffer
.Start:
    cmp [NoteBufferEnd],esi     ;beyond end of buffer?
    jbe .EndLoop
    cmp [NoteBuffer+esi],edx    ;reached desired time yet?
    jb .Next
.EndLoop:

    ; flag change for any channels enabled or disabled by user since last call
    ; channels on     = channels changed & channels used
    ; new channels   |= ((channels on & channels enabled)
    ;                   ^ channels previously active)
    ; channels active = channels on & channels enabled
    ;!!!
    mov eax,[edi+SeekNoteStruct.Change]
    mov ebx,[edi+SeekNoteStruct.Used]   ;get which channels are on/off
    and eax,ebx                         ;get channels turned on
    ;and ebx,[MutedVoices]              ;turn off any user disabled channels
    mov [edi+SeekNoteStruct.Pos],esi    ;save new position in buffer
    mov ecx,ebx
    ;xor ebx,[edi+SeekNoteStruct.Active]
    mov [edi+SeekNoteStruct.Active],ecx
    mov [edi+SeekNoteStruct.On],eax
    ;mov [edi+SeekNoteStruct.Change],eax
    ;or [edi+SeekNoteStruct.Change],ebx

    ;(eax=notes turned on)
    ;mov eax,[edi+SeekNoteStruct.Change]
    xor ecx,ecx
    jmp short .FirstVoice
.NextVoice:
    mov edx,[edi+SeekNoteStruct.Atrs+ecx*4]  ;get voice/sample number/pitch
    shld ebx,edx,16+2
    and ebx,1111111100b         ;make ptr of sample number
    mov ebx,[BrrTable.AtrsIdx+ebx]
    sub dh,byte 207             ;reduce note value for 32khz playrate to 0
    add dh,[SampleList.HalfSemitone+ebx]
    shr dh,1                    ;note number = half-semitone / 2
    mov [edi+SeekNoteStruct.Pitches+ecx],dh
    ;mov [edi+SeekNoteStruct.Pitches+ecx],byte 0
.SkipVoice:
    inc ecx
.FirstVoice:
    shr eax,1
    jc .NextVoice
    jnz .SkipVoice

    add esp,byte 4              ;discard flags for channels turned off
    ret                         ;(esi=new position in buffer)

;<SOUND.ASM> 
    mov esi,[NoteBufferPlayed]
    cmp [NoteBufferEnd],esi
    jbe near .End2
.NextNote:
    mov eax,[NoteBuffer+esi]    ;get cycle time when note was recorded
    cmp [TickTime],eax
    jb near .End
    movzx eax,byte [NoteBuffer+esi+4]   ;get voice number
.KeyOn:
    test al,15
    jnz .KeyOff
    shr eax,4                   ;select voice
    movzx ebx,byte [NoteBuffer+esi+6] ;get sample number
    bt dword [ActiveChannels],eax
    jnc .AfterKeyChange
    mov edi,[BrrTable.AtrsIdx+ebx*4]
    ;shl edi,SampleList.Shift
    mov bl,[SampleList.Instrument+edi]
    cmp [Sound.UsedFmPatches+eax*4],bl
    je .SameInstrument
    mov [Sound.UsedFmPatches+eax*4],bl
    push eax                    ;save voice
    push esi                    ;save pointer in note buffer
    shl ebx,4                   ;*16 get instrument information
    lea esi,[SbFm.PatchData+ebx];get ptr to instrument information in patch table
    call SetFmPatch
    pop esi
    pop eax
.SameInstrument:
    call SoundNoteOn
    jmp short .AfterKeyChange
    ;jmp short .DebugPause
.KeyOff:
    test al,14
    jnz .KeyPitchChange
    shr eax,4                   ;select voice
    bt dword [ActiveChannels],eax
    jnc .AfterKeyChange
    call SoundNoteOff
    ;jmp short .AfterKeyChange
.KeyPitchChange:
    ;call something...
.AfterKeyChange:
    add esi,byte 8
    cmp [NoteBufferEnd],esi     ;ptr < NoteBufferSize
    ja near .NextNote

.End:
    mov [NoteBufferPlayed],esi
.End2:
    ret


;====================
; Turns a channel on (using the current patch for that channel)
;
;(esi=ptr in note buffer, eax=voice)
SoundNoteOn:
    bts dword [Sound.ChannelsOn],eax    ;flag that voice was turned on recently
    btr dword [Sound.ChannelsOff],eax   ;clear that voice needs to be turned off
    bts dword [Sound.ChannelsUsed],eax  ;flag voice is in use
    mov edi,eax                 ;make copy of voice number
    jnc .AlreadyOff
    add al,0B0h                 ;DSP register, low byte of pitch
    ;xor ah,ah                   ;turn voice off
    call WriteSbReg
.AlreadyOff:

    mov dl,[NoteBuffer+esi+5]   ;get nearest half-semitone
    movzx ebx,byte [NoteBuffer+esi+6] ;get sample number
    mov ebx,[BrrTable.AtrsIdx+ebx*4]
    sub edx,207                 ;reduce playrate to centerpoint, for any samples at 32khz (~4096 rate)
    ;shl ebx,SampleList.Shift
    add dl,[SampleList.HalfSemitone+ebx]
    and edx,11111110b           ;get MIDI note number (*2)
    lea eax,[edi+0A0h]          ;DSP register, low byte of pitch
    mov ebx,[SbFreqTbl+edx]     ;get fnumber for note's pitch
    mov ah,bl
    call WriteSbReg
    lea eax,[edi+20B0h]         ;DSP register, high byte of pitch (20h turns voice on)
    mov [Sound.UsedFnumbers+edi],bh ;save fnumber/octave to eliminate click when turning voice off
    or ah,bh
    jmp WriteSbReg
;.End:
;    ret

;====================
;(esi=ptr in note buffer, eax=voice)
SoundNoteOff:
    bt dword [Sound.ChannelsOn],eax     ;was voice recently turned on
    jc .Later                           ;if so, turn if off the next call
.Now:
    btr dword [Sound.ChannelsUsed],eax  ;clear voice as being used
    mov ah,[Sound.UsedFnumbers+eax]     ;get previous fnumber/octave of voice, to prevent clicking sound
    add al,0B0h                         ;DSP register, low byte of pitch
    ;xor ah,ah                          ;turn voice off, as long as bit 20h is clear
    jmp WriteSbReg
.Later:
    bts dword [Sound.ChannelsOff],eax   ;simply flag voice as needing to be turned off
    ret

;<DSPEMU.ASM>
;GetSampleWaveAttributes:
; My dream function that awaits completion! :)
; Not sure exactly how I'll get it to work... :(
;
; Scans from the start point to the end of the sample (usually starting at
; the loop point or beginning of the sample if there is none) to return the
; dominant frequency and average volume.
;
;  1                  Ŀ    Ŀ
;  0    
; -1                           
;    L R    LEFT - RIGHT = TOTAL
;    ٳ        4 - -2 =  6    2+6+8 =  16  close
;     ٳ        4 - -4 =  8
;      ٳ        2 - -4 =  6
;       ٳ        0 - -2 =  2
;        ٳ       -2 -  0 = -2
;                -4 -  2 = -6
;               5 - -5 = 10    2+6+10 = 18  perfect resonance
;               3 - -3 =  6
;               1 - -1 =  2
;              -1 -  1 = -2
;              -3 -  3 = -6
;              -5 -  5 = 10
;             4 - -2 =  6    2+6+8 =  16  close
;             2 -  0 =  2
;             0 - -2 = -2
;            -2 -  4 = -6
;            -4 -  4 = -8
;            -4 -  2 = -6
;     0 -  0 =  0    0+0+0 = 0    complete cancellation
;   ret

<SPC2MIDI.ASM>
;
; FM attributes for sound effects.
SoundFxList:
.Total              equ 300
.eSize              equ 64
.Shift              equ 4
.MultMod            resb 1  ;amp mod/vib/eg/key scale/multiple
.MultCar            resb 1
.LevelMod           resb 1  ;key scale level/volume
.LevelCar           resb 1
.AtkDecMod          resb 1  ;attack rate/decay rate
.AtkDecCar          resb 1
.StnRlsMod          resb 1  ;sustain level/release rate
.StnRlsCar          resb 1
.FormMod            resb 1  ;waveform
.FormCar            resb 1
.FeedBack           resb 1  ;feedback level/connection
.Name:              resb .eSize+$-SoundFxList
.NameSize           equ 30
                    resb ($-SoundFxList)*(.Total-1)
.Used:              resb .Total/8 ;bit flags to indicate which entries contain information


;<DSPEMU.ASM>
BrrSampleExpand: ;(no filters)
...
.NextByte:
    mov dl,byte [SpcRam+esi]    ;get high and low nybble
    mov ebx,edx
    shl edx,28                  ;move low nybble to top of dword
    shl ebx,24                  ;move high nybble to top of dword
    sar edx,cl                  ;apply range to low nybble
    sar ebx,cl                  ;apply range to high nybble
    mov [edi],bx                ;high nybble is stored
    mov [edi+2],dx              ;before low nybble
    inc esi
    add edi,byte 4
    dec ch
    jg .NextByte

<DSPEMU.ASM>
; To debug the sample processing
    mov ecx,Screen.Height
    mov edi,Screen.Adr+(Screen.Width/2)
    mov esi,[.Ptr]
    mov edx,[.Acm]
    mov ebx,[.Mod]

    lea eax,[SamplesBuffer+esi*2]
.NextSample:
    ; Get sample and advance wave by playrate
    movsx eax,word [SamplesBuffer+esi*2]
    add edx,ebx
    adc esi,[.Inc]

    ; Merge voice into buffer
    ; (eax=sample value, edi=mix buffer ptr)
    xchg ah,[.Cols+ecx-1]
    movsx eax,ah
    mov byte [edi+eax],0
    movsx eax,byte [.Cols+ecx-1]
    mov byte [edi+eax],15

    ; Check for end of sample
    cmp [.End],esi
    jle .SampleEnd
.NotEnd:
    add edi,Screen.Width

    dec ecx
    jg .NextSample

    mov [.Ptr],esi
    mov [.Acm],edx
    mov [.Mod],ebx
    ret


;<DSPEMU.ASM>
BrrSampleExpand:
    ; note this routine assumes the sample has already been checked for
    ; validity!
    mov dword [.PrevSample],0   ;previous sample before current one
    mov dword [.PrevSample2],0  ;sample two back from current one

    ; calculate desination in buffer (sample_offset / 9 * 32)
    mov eax,esi                 ;for the division
    mov ebx,9
    xor edx,edx
    push esi                    ;save sample base for later
    div ebx                     ;/9
    push dword 0                ;set sample length counter to zero
    shl eax,5                   ;*32
    lea edi,[SamplesBuffer+eax]
    xor eax,eax
    align 16

.Next:
;(eax<=255!, esi=brr source, edi=buffer dest)
    mov al,[SpcRam+esi]         ;get header byte
    inc esi
    mov ecx,eax                 ;copy header byte
    and eax,15                  ;isolate filter and end bits only
    shr ecx,4                   ;get range from top four bits
    not ecx
    add ecx,29|(8<<8)           ;get 28-range  (!range+29)
                                ;and set counter for eight samples to follow
.NextByte:
    mov dl,byte [SpcRam+esi]    ;get high nybble
    and dl,0F0h                 ;mask out low nybble
    shl edx,24                  ;move high nybble to top of dword
    sar edx,cl                  ;apply range to high nybble

    ; apply filter to high nybble
    cmp eax,1000b
    jb .NoSecondFilter1
    mov ebx,[.PrevSample2]
    imul ebx,[.FilterTbl2+eax*4]
    sar ebx,8                   ;/256
    sub edx,ebx
.NoSecondFilter1:
    mov ebx,[.PrevSample]
    mov [.PrevSample2],ebx
    cmp eax,100b
    jb .NoFilter1
    imul ebx,[.FilterTbl+eax*4]
    sar ebx,8                   ;/256
    add edx,ebx
.NoFilter1:
    mov [.PrevSample],edx
    mov [edi],dx                ;store high nybble first

    mov bl,byte [SpcRam+esi]    ;get low nybble
    shl ebx,28                  ;move low nybble to top of dword
    sar ebx,cl                  ;apply range to low nybble

    ; apply filter to low nybble
    cmp eax,1000b
    jb .NoSecondFilter2
    mov edx,[.PrevSample2]
    imul edx,[.FilterTbl2+eax*4]
    sar edx,8                   ;/256
    sub ebx,edx
.NoSecondFilter2:
    mov edx,[.PrevSample]
    mov [.PrevSample2],edx
    cmp eax,100b
    jb .NoFilter2
    imul edx,[.FilterTbl+eax*4]
    sar edx,8                   ;/256
    add ebx,edx
.NoFilter2:
    mov [.PrevSample],ebx
    mov [edi+2],bx              ;store low nybble last

    inc esi
    add edi,byte 4
    dec ch
    jg near .NextByte

    add dword [esp],byte 16     ;sample count+=16
    test eax,1
    jnz .BrrEnd
    cmp esi,65536-9
    jbe near .Next
.BrrEnd:

    mov ecx,esi
    pop edx                     ;retrieve length in samples
    pop esi                     ;retrieve sample base
    sub ecx,esi                 ;[OR] byte_length = starting_offset - ending_offset
    ret


section .bss
alignb 4
.PrevSample:    resd 1          ;previous sample before current one
.PrevSample2:   resd 1          ;sample two back from current one
section .data
align 4
; All filter ratios are scaled up to a denominator of 256
; 15/16=240/256  61/32=488/256  115/64=460  13/16=208/256
.FilterTbl equ $-16             ;don't store any filter values for filter 0
    dd 240,240,240,240, 488,488,488,488, 460,460,460,460
.FilterTbl2 equ $-32            ;don't store any filter values for filter 0&1
    dd 240,240,240,240, 208,208,208,208
section .text


<SOUND.ASM>
;
; Copies mixing buffer to sound card buffer. Ensures the output and DMA are
; synchronized, in you switch away to Windows, or in case the sound card does
; not have accurate timing (which I have found!). It tries to always stay
; one interrupt's buffer length ahead of the sound card.
;
; () ()
CopyMixBuffer: #1
    call GetDmaLag
%if 1
    mov [Sound.DmaPos],eax
    neg eax
    mov edx,[Sound.BufferPos]
    ;mov [Sound.BufferPre],edx
    add eax,edx
    and eax,Sound.BufferSize-1
    mov edx,[Sound.BufferDif]
    mov [Sound.BufferDif],eax
    sub eax,edx
    mov [Sound.BufferPre],eax
%endif

    mov edi,[Sound.BufferPtr]
    mov edx,[Sound.BufferPos]
    mov esi,MixBuffer
    mov ecx,[Sound.MixSamples]
    add edi,edx

    ; resync if necessary
    ;test byte [SoundFlags],SoundFlags.Resync
    ;jz

    ; check for buffer and wrap and transfer from mix buffer to sound buffer
    test byte [SoundFlags],SoundFlags.16bit
    jz .8bit
    shr edx,1                   ;/2 for words
    add edx,ecx                 ;buffer offset + number of mixed samples
    sub edx,(Sound.BufferSize/2) ;buffer end > buffer size?
    jl .16bitNoWrap
    sub ecx,edx
    call TransferWave16bit
    mov ecx,edx
    sub edi,Sound.BufferSize
    ;mov edi,[Sound.BufferPtr]
.16bitNoWrap:
    call TransferWave16bit
    jmp short .TransferDone
.8bit:
    add edx,ecx                 ;buffer offset + number of mixed samples
    sub edx,Sound.BufferSize    ;buffer end > buffer size?
    jl .8bitNoWrap
    sub ecx,edx
    call TransferWave8bit
    mov ecx,edx
    sub edi,Sound.BufferSize
    ;mov edi,[Sound.BufferPtr]
.8bitNoWrap:
    call TransferWave8bit
.TransferDone:

    sub edi,[Sound.BufferPtr]
    and edi,Sound.BufferSize-1
    mov [Sound.BufferPos],edi
    ret


<SOUND.ASM>
;
; Copies mixing buffer to sound card buffer. Ensures the output and DMA are
; synchronized, in case you switch away to Windows, or in if the sound card
; does not have accurate timing (which I have found!). It tries to always
; stay one interrupt's buffer length or more ahead of the sound card.
;
; () ()
CopyMixBuffer: #2
    push ebp
    call GetDmaLag              ; BufferPos - DmaPos should be IntSamples
    mov [Sound.BufferDif],eax   ;*** debug ***

    ; check deviation difference
    ; Unnacceptable deviation is <=IntSamples*.5 | >=IntSamples*1.5
    ; Allowable deviation is >=IntSamples*(1/16)
    ;mov ecx,[Sound.MixSamples]
    mov ecx,[Sound.IntSamples]
    mov edx,ecx
    sub eax,ecx                 ;deviation = (buffer pos - DMA pos) - samples per int
    shr edx,1                   ;/2
    cmp eax,edx
    jbe .Synced                 ;difference < (samples per int / 2) unsigned
    jge .WayOff                 ;difference >= (samples per int / 2)
    neg eax
    cmp eax,edx                 ;-difference >= (samples per int / 2)
    jge .WayOff
    ;jmp short .Synced

.OffSync:
    ; Say we are playing at 22khz and are 80 samples behind, here's how to
    ; scale the contents of the MixBuffer to the output buffer.
    ;   22050hz / 30 frames per second = 735 samples per interrupt
    ;   735 / 16 minumum deviation percentage = 45 minimum deviation samples
    ;   80 samples behind - 45 minimum deviation = 35
    ;   35 + 735 samples per int = 770 samples to catch up
    ;   735 / 770 = .954 ratio
    ; Since decimal points are unuseable, shift up before the division.
    ; The maximum samples per interrupt (playrate of 64khz) is 2134, so shift
    ; up by the max amount that will keep it in 32 bits. 2^20
    ;   735 << 20 = 770703360
    ;   770703360 / 770 = 1000913
    ;   1000913 << (32-20) to put it at the very top of the dword
    ;   4099739648 / 4294967296
     ; Now every time the numerator is added and overflows, carry will be set
    ; and the source ptr will advance by one.

%if 0
    mov eax,[Sound.MixSamples]
    mov ecx,[Sound.IntSamples]
    inc ecx
    mov [Sound.IntSamples],ecx
    shl eax,20                  ;*1048576
    xor edx,edx
    div ecx                     ;samples per int / sb samples
    shl eax,12                  ;move bit 19 up to bit 31
    mov edx,eax
    jmp short .CopyBuffer

%else
    xchg eax,ecx                ;swap samples per int with samples behind
    ;xor eax,eax
    add ecx,eax                 ;sb samples = samples behind + samples per int
    shl eax,20                  ;*1048576
    xor edx,edx
    div ecx                     ;samples per int / sb samples
    shl eax,12                  ;move bit 19 up to bit 31
    mov edx,eax
    dec edx
    jmp short .CopyBuffer
%endif

.WayOff:
    ;inc dword [Sound.BufferPre] ;*** debug ***
    mov eax,[Sound.DmaPos]
    add eax,ecx                 ;+=Sound.MixSamples
    and eax,Sound.BufferSize-2  ;wrap within buffer to nearest word
    mov [Sound.BufferPos],eax

.Synced:
    mov edx,-1

; (ecx=buffer sample count, edx=modulus increment)
.CopyBuffer:
    ;mov esi,MixBuffer
    mov ebp,[Sound.DmaMod]      ;remainder accumulator
    mov ebx,[Sound.BufferPtr]
    mov edi,[Sound.BufferPos]
    ;shr esi,2                   ;/4
    xor esi,esi

    test byte [SoundFlags],SoundFlags.16bit
    jz .8bitNext

.16bitNext:
    and edi,Sound.BufferSize-2  ;wrap (to even word)
    mov eax,[esi*4]
    add ebp,edx
    adc esi,0
    cmp eax,32768
    jb .16bitValid
    cmp eax,-32768
    jae .16bitValid
    sar eax,31                  ;sign extend
    xor eax,32767
.16bitValid:
    mov [ebx+edi],ax
    add edi,byte 2
    dec ecx
    jg .16bitNext
    jmp short .End

.8bitNext:
    and edi,Sound.BufferSize-1  ;wrap
    mov eax,[esi*4+MixBuffer]
    cmp esi,[Sound.IntSamples]
    jb .ok
    inc dword [Sound.BufferPre] ;*** debug ***
.ok:
    add ebp,edx
    adc esi,0
    add eax,32768               ;toggle sign bit for conversion to 8bit
    test eax,0FFFF0000h
    jz .8bitValid
    sar eax,31                  ;sign extend
    not eax
.8bitValid:
    mov [ebx+edi],ah
    inc edi
    dec ecx
    jg .8bitNext

.End:
    mov [Sound.DmaMod],ebp
    and edi,Sound.BufferSize-1
    mov [Sound.BufferPos],edi
    pop ebp
    ret


;
<SOUND.ASM>
CopyMixBuffer #3

    push ebp
    call GetDmaLag
    mov dword [SoundDebug.Lag],eax

    ; check deviation difference
    ; BufferPos - DmaPos should be MixSamples
    ; unnacceptable deviation is <=IntSamples*.5 | >=IntSamples*1.5
    mov edx,[Sound.MixSamples]
    mov ecx,edx
    jmp short .Synced ;***
    shr edx,1                   ;/2 half samples per interrupt
    sub eax,edx                 ;deviation = lag - (samples per int / 2)
    jle .OffSync                ;difference <= (samples per int / 2)
    cmp eax,ecx
    jb .Synced                  ;difference >= (samples per int / 2)

.OffSync:
    ; resync buffer position ahead of what DMA is currently reading
    ;mov edi,[Sound.BufferPos]

    mov edi,[Sound.DmaPos]
    add edi,ecx                 ;+=Sound.MixSamples
    and edi,Sound.BufferSize-2 ;wrap within buffer to nearest word
    inc dword [SoundDebug.Counter]
    jmp short .CopyBuffer

; (ecx=mixed samples to copy,)
.Synced:
    shr dword [Sound.MixedSamples],1
    shr dword [Sound.DmaSamples],1
    add [Sound.MixedSamples],ecx
    mov edi,[Sound.BufferPos]
    add [Sound.DmaSamples],ebx

; (ecx=mixed samples to copy, edi=buffer position)
.CopyBuffer:
    mov esi,MixBuffer
    mov ebp,[Sound.DmaSamples]
    mov ebx,[Sound.BufferPtr]
    mov edx,[Sound.DmaMod]      ;remainder accumulator
    sub ebp,[Sound.MixedSamples]  ;numerator = DMA samples - mixed samples
    mov dword [SoundDebug.DifSamples],ebp ;***

    ;***
    ;xor edx,edx
    mov dword [SoundDebug.OutSamples],0
    ;xor ebp,ebp
    ;mov ebp,[SoundDebug.SampleInc]
    ...


;
<SOUND.ASM>
SyncToDma:
    ; read address from 8237 DMA controller (weird configuration)
    mov edx,[Sound.DmaChannel]
    xor eax,eax                 ;zero address
    cmp dl,4
    jae .Hdma
    mov ecx,eax                 ;zero left shift
    out 0Ch,al                  ;clear flip-flop (value is irrelavant)
    shl edx,1                   ;*2 address port
    jmp short .ReadAddress
.Hdma:
    lea edx,[0C0h+edx*4-16]     ;address port
    out 0D8h,al                 ;clear flip-flop (value is irrelavant)
    mov cl,1                    ;set left shift (address*2)
.ReadAddress:
    in al,dx                    ;get address low byte
    mov ah,al
    in al,dx                    ;get address high byte
    xchg al,ah                  ;correct high/low byte order

    ; get offset in buffer & calculate lag
    shl eax,cl                  ;*2 for 16bit channel, *1 for 8bit channel
    ;mov edx,[Sound.IntSamples]
    mov edx,[Sound.MixSamples]
    sub eax,[Sound.BufferPtr]   ;buffer offset = DMA address - base ptr
    mov ecx,edx
    shr ecx,2                   ;/4
    mov ebx,[Sound.BufferPos]
    add eax,ecx                 ;+25% hack for Crystal Audio
    and eax,Sound.BufferSize-1
    sub ebx,eax                 ;lag = buffer pos - trailing DMA pos
    ; this does not use cl from above because some sound cards utilize low
    ; DMA channels for 16bit data
    and ebx,Sound.BufferSize-1
    test byte [SoundFlags],SoundFlags.Force8bit
    jnz .8bit
    shr ebx,1                   ;lag /2 for words
.8bit:

    ;mov [PrintDebugInfo.Value1],eax
    ;mov [PrintDebugInfo.Value2],ebx

    ; check deviation difference (BufferPos - DmaPos should be >=IntSamples)
    ; (eax=DMA buffer offset, ebx=lag, edx=int samples)
    sub ebx,edx                 ;lag < buffer size
    jb .Behind                  ;we're falling behind, DMA will soon beat us
    sub ebx,edx                 ;lag < (buffer size * 2)
    jb .InSync                  ;just fine
    cmp ebx,edx                 ;we're faster than DMA
    jb .Ahead

; somehow got way off (Alt+Tab probably to blame)
; (eax=DMA buffer offset, edx=buffer size)
.OffSync:
    ; resync buffer position ahead of what DMA is currently reading
    lea ebx,[edx+edx*2]         ;Sound.IntSamples*3
    test byte [SoundFlags],SoundFlags.Force8bit
    jnz .8bitResync
    shl ebx,1                   ;*2 if words, else bytes
.8bitResync:
    shr ebx,1                   ;/2  (*3/2 = 1.5)
    add ebx,eax                 ;(Sound.IntSamples*1.5)+Dma address
    and ebx,Sound.BufferSize-2  ;wrap within buffer to nearest word
    mov [Sound.BufferPos],ebx
.InSync:
    ret

; (edx=buffer size)
.Behind:
    lea edx,[edx+edx*2]         ;mix 1.5 buffers
.Ahead:
    shr edx,1                   ;mix .5 buffers
    ret
